@@ -6,23 +6,22 @@
 
 /* Includes. ================================================================*/
 
-#include <ctype.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <stdarg.h>
 #include <string.h>
 #include <setjmp.h>
-#include <locale.h>
+#include <glib.h>
 #include "basic.h"
 
 /* Global variables. ========================================================*/
 
-char_t malaga_version[] = "6.13";
+char_t malaga_version[] = "7.8";
 string_t program_name;
 bool_t in_emacs_malaga_mode;
-bool_t hangul;
+bool_t split_hangul_syllables;
 text_t *error_text;
-jmp_buf *_error_handler;
+jmp_buf *current_error_handler;
 bool_t user_break_requested;
 
 /* List functions. ==========================================================*/
@@ -316,9 +315,29 @@ add_char_to_text( text_t *text, char_t character )
 
 /*---------------------------------------------------------------------------*/
 
+void 
+add_unichar_to_text( text_t *text, gunichar c )
+/* Add C to TEXT. */
+{ 
+  char buf[7];
+  int_t n;
+
+  n = g_unichar_to_utf8( c, buf );
+  buf[n] = EOS;
+  if (text->buffer_size < text->string_size + n + 1) 
+  { 
+    text->buffer_size = renew_vector( &text->buffer, sizeof( char_t ), 
+                                      2 * (text->string_size + n) );
+  }
+  strcpy( text->buffer + text->string_size, buf );
+  text->string_size += n;
+}
+
+/*---------------------------------------------------------------------------*/
+
 void 
 insert_in_text( text_t *text, string_t string, int_t position )
-/* Insert STRING at POSITION in TEXT (position counts from 0 onward). */
+/* Insert STRING at POSITION in TEXT (position counts bytes from 0 onward). */
 { 
   int_t string_len;
 
@@ -361,12 +380,12 @@ insert_char_in_text( text_t *text, char_t character, int_t position )
 
 /*---------------------------------------------------------------------------*/
 
-string_t 
+char_t * 
 text_to_string( text_t **text_p )
 /* Return content of *TEXT_P as a string and delete *TEXT_P.
  * The string must be freed after use. */
 {
-  string_t string;
+  char_t *string;
 
   string = new_string( (*text_p)->buffer, NULL );
   free_text( text_p );
@@ -377,10 +396,10 @@ text_to_string( text_t **text_p )
 
 /* Add string_t FORMAT to text_t TEXT, replace any "%s" with the next
  * "string_t" argument in va_list ARGS, any "%c" with the next "char" argument,
- * and any "%d" with the next "int_t" argument. */
+ * any "%u" with the next "gunichar" argument, and any "%d" with the next 
+ * "int_t" argument. */
 #define VA_PRINT_TEXT( text, format, args ) \
 do { \
-  int_t pos, num; \
   string_t format_p; \
   \
   for (format_p = format; *format_p != EOS; format_p++) \
@@ -393,10 +412,23 @@ do { \
       case 'c': \
         add_char_to_text( text, va_arg( args, int ) ); \
         break; \
+      case 'u': \
+      { \
+        char buf[7]; \
+        int n; \
+        \
+        n = g_unichar_to_utf8( va_arg( args, gunichar ), buf); \
+        buf[n] = EOS; \
+	add_to_text( text, buf ); \
+	break; \
+      } \
       case 's': \
         add_to_text( text, va_arg( args, string_t ) ); \
         break; \
       case 'd': \
+      { \
+        int_t pos, num; \
+        \
         num = va_arg( args, int_t ); \
         if (num < 0) \
         { \
@@ -410,6 +442,7 @@ do { \
           num /= 10; \
         } while (num != 0); \
         break; \
+      } \
       default: \
         add_char_to_text( text, *format_p ); \
         break; \
@@ -422,13 +455,14 @@ do { \
 
 /*---------------------------------------------------------------------------*/
 
-void 
+void
 print_text( text_t *text, string_t format, ... )
 /* Add FORMAT to TEXT, replace any "%s" with a "string_t" argument, any "%c"
- * with a "char" argument, and any "%d" with an "int_t" argument. */
-{ 
+ * with a "char" argument, any "%u" with a "gunichar" argument, and any "%d"
+ * with an "int_t" argument. */
+{
   va_list args;
-  
+
   va_start( args, format );
   VA_PRINT_TEXT( text, format, args );
   va_end( args );
@@ -436,17 +470,18 @@ print_text( text_t *text, string_t format, ... )
 
 /* String functions. ========================================================*/
 
-string_t 
+char_t * 
 new_string( string_t string, string_t end )
 /* Allocate memory and copy STRING into it.
  * If END != NULL, it marks the end of the string.
  * The result string must be freed after use. */
-{ 
-  string_t new_str, new_str_p;
+{
+  char_t *new_str;
+  char_t *new_str_p;
 
   if (end == NULL) 
     end = string + strlen( string );
-  new_str_p = new_str = new_vector( sizeof( char_t ), end - string + 1 );
+  new_str = new_str_p = new_vector( sizeof( char_t ), end - string + 1 );
   while (string < end) 
     *new_str_p++ = *string++;
   *new_str_p = EOS;
@@ -455,42 +490,44 @@ new_string( string_t string, string_t end )
 
 /*---------------------------------------------------------------------------*/
 
-string_t 
+char_t * 
 new_string_readable( string_t from, string_t from_end )
 /* Like "new_string", but enclose the string in double quotes, copy a "\" in 
- * front of quotes and copy any control chars in octal code: "\000". 
+ * front of quotes and backslashed, and copy control chars in "\uxxxx" format.
  * If FROM_END != NULL, it marks the end of the string. 
  * The result string must be freed after use. */
 {
   text_t *text;
-  int_t i, code, position, digit;
+  int_t i, code, position;
 
   text = new_text();
   if (from_end == NULL) 
     from_end = from + strlen( from );
   add_char_to_text( text, '\"' );
-  for (; from < from_end; from++) 
+  while (from < from_end)
   { 
     if (*from == '\"' || *from == '\\')  /* Prepend a backslash. */
     { 
       add_char_to_text( text, '\\' );
-      add_char_to_text( text, *from );
+      add_char_to_text( text, *from++ );
     } 
-    else if (IS_PRINT( *from )) 
-      ADD_CHAR_TO_TEXT( text, *from );
-    else /* Convert control chars to backslash and hex code. */
+    else if ((*from >= 0 && *from < 32) || *from == 127)
     { 
-      add_to_text( text, "\\x" );
+      /* Convert control chars to octal "\xxx" format. */
+      add_char_to_text( text, '\\' );
       position = text->string_size;
-      code = ORD( *from );
-      for (i = 0; i < 2 * (int_t) sizeof( char_t ); i++) 
+      code = *from++;
+      for (i = 0; i < 3; i++) 
       { 
-	digit = code % 16;
-	digit += (digit < 10 ? '0' : 'a' - 10);
-	insert_char_in_text( text, digit, position );
-        code = code / 16;  
+	insert_char_in_text( text, code % 8 + '0', position );
+        code = code / 8;  
       }
     } 
+    else
+    {
+      add_unichar_to_text( text, g_utf8_get_char( from ) );
+      from = g_utf8_next_char( from );
+    }
   }
   add_char_to_text( text, '\"' );
   return text_to_string( &text );
@@ -498,7 +535,7 @@ new_string_readable( string_t from, string_t from_end )
 
 /*---------------------------------------------------------------------------*/
 
-string_t 
+char_t *
 concat_strings( string_t first_string, ... )
 /* Concatenate a list of strings and return the result string.
  * Must have NULL-terminated list of strings as parameters.
@@ -506,7 +543,9 @@ concat_strings( string_t first_string, ... )
 {
   va_list args;
   size_t length;
-  string_t next_string, string, string_p;
+  string_t next_string;
+  char_t *string;
+  char_t *string_p;
 
   /* Compute length of the result string. */
   va_start( args, first_string );
@@ -542,8 +581,8 @@ string_t
 next_non_space( string_t string )
 /* Return STRING, but without leading spaces. */
 { 
-  while (IS_SPACE( *string )) 
-    string++;
+  while (g_unichar_isspace( g_utf8_get_char( string ) ))
+    string = g_utf8_next_char( string );
   return string;
 }
 
@@ -556,16 +595,24 @@ strcmp_no_case( string_t str1, string_t str2 )
  *            0 if STR1 = STR2,
  *            1 if STR1 > STR2. */
 { 
-  /* Find first char where STR1 and STR2 differ. */
-  while (TO_LOWER( *str1 ) == TO_LOWER( *str2 )) 
-  { 
-    if (*str1 == EOS) 
-      return 0;
-    str1++;
-    str2++;
-  } 
+  gunichar c1, c2;
 
-  return (ORD( TO_LOWER( *str1 ) ) < ORD( TO_LOWER( *str2 ) )) ? -1 : 1;
+  /* Find first char where STR1 and STR2 differ. */
+  while (TRUE)
+  {
+    if (*str1 == EOS)
+      return (*str2 == EOS ? 0 : -1);
+    else if (*str2 == EOS)
+      return 1;
+
+    c1 = g_unichar_tolower( g_utf8_get_char( str1 ));
+    c2 = g_unichar_tolower( g_utf8_get_char( str2 ));
+    if (c1 != c2)
+      return (c1 < c2 ? -1 : 1);
+
+    str1 = g_utf8_next_char( str1 );
+    str2 = g_utf8_next_char( str2 );
+  } 
 }
 
 /*---------------------------------------------------------------------------*/
@@ -573,29 +620,38 @@ strcmp_no_case( string_t str1, string_t str2 )
 int_t 
 strncmp_no_case( string_t str1, string_t str2, int_t n )
 /* Return (case insensitive) lexical order of STR1 and STR2,
- * but compare only the first N characters.
+ * but compare only the first N bytes.
  * Result is -1 if STR1 < STR2,
  *            0 if STR1 = STR2,
  *            1 if STR1 > STR2. */ 
 { 
-  /* Find first char where STR1 and STR2 differ. */
-  while (TO_LOWER( *str1 ) == TO_LOWER( *str2 ) && n > 0)  
-  { 
-    if (*str1 == EOS) 
-      return 0;
-    str1++;
-    str2++;
-    n--;
-  }
+  string_t str1_end = str1 + n;
+  gunichar c1, c2;
 
-  if (n == 0) 
-    return 0;
-  return (ORD( TO_LOWER( *str1 ) ) < ORD( TO_LOWER( *str2 ) )) ? -1 : 1;
+  /* Find first char where STR1 and STR2 differ. */
+  while (TRUE)
+  {
+    if (str1 == str1_end)
+      return 0;
+    else if (*str1 == EOS)
+      return (*str2 == EOS ? 0 : -1);
+    else if (*str2 == EOS)
+      return 1;
+
+    c1 = g_unichar_tolower( g_utf8_get_char( str1 ));
+    c2 = g_unichar_tolower( g_utf8_get_char( str2 ));
+
+    if (c1 != c2)
+      return (c1 < c2 ? -1 : 1);
+
+    str1 = g_utf8_next_char( str1 );
+    str2 = g_utf8_next_char( str2 );
+  }
 }
 
 /*---------------------------------------------------------------------------*/
 
-string_t 
+char_t *
 replace_arguments( string_t format, string_t chars, ... )
 /* Create a new string with a copy of FORMAT.
  * Replace each sequence "%C" in FORMAT, where C is the N-th
@@ -645,7 +701,7 @@ replace_arguments( string_t format, string_t chars, ... )
 
 /*---------------------------------------------------------------------------*/
 
-string_t 
+char_t *
 double_to_string( double number )
 /* Convert NUMBER to a string. It must be freed after use. */
 { 
@@ -657,7 +713,7 @@ double_to_string( double number )
 
 /*---------------------------------------------------------------------------*/
 
-string_t 
+char_t * 
 int_to_string( int_t number )
 /* Convert NUMBER to a string. It must be freed after use. */
 {
@@ -670,12 +726,12 @@ int_to_string( int_t number )
 /* Error handling. ==========================================================*/
 
 void 
-_throw( void )
+throw( void )
 /* Call the current error handler. 
  * If there is no current error handler, print error and exit. */
 { 
-  if (_error_handler != NULL) 
-    longjmp( *_error_handler, 1 );
+  if (current_error_handler != NULL) 
+    longjmp( *current_error_handler, 1 );
   else 
   { 
     fprintf( stderr, "%s: %s\n", program_name, error_text->buffer );
@@ -689,7 +745,7 @@ void
 complain( string_t message, ... )
 /* Save the error MESSAGE in ERROR_TEXT.
  * Additional arguments to "complain" are inserted where 
- * "%s" (string_t ARGUMENT), "%c" (char_t ARGUMENT)
+ * "%s" (string_t ARGUMENT), "%c" (char_t ARGUMENT), "%u" (gunichar ARGUMENT),
  * or "%d" (int_t ARGUMENT) is part of MESSAGE. */
 { 
   va_list args;
@@ -698,7 +754,7 @@ complain( string_t message, ... )
   va_start( args, message ); 
   VA_PRINT_TEXT( error_text, message, args );
   va_end( args );
-  _throw();
+  throw();
 }
 
 /*---------------------------------------------------------------------------*/
@@ -732,7 +788,6 @@ init_basic( string_t prog_name )
     setbuf( stderr, NULL );
   }
 #endif
-  setlocale( LC_CTYPE, "" );
 }
 
 /*---------------------------------------------------------------------------*/
