@@ -13,6 +13,7 @@
 #include <stdlib.h>
 #include <setjmp.h>
 #include <time.h>
+#include <glib.h>
 #include "basic.h"
 #include "pools.h"
 #include "values.h"
@@ -23,6 +24,7 @@
 #include "analysis.h"
 #include "input.h"
 #include "commands.h"
+#include "commands_interactive.h"
 #include "options.h"
 #include "display.h"
 #include "malaga_lib.h"
@@ -40,23 +42,25 @@
 
 /* Variables. ===============================================================*/
 
-static string_t analysis_input; /* Input line for interactive analysis. */
+static char_t *analysis_input; /* Input line for interactive analysis. */
 
-static int_t debug_analysis_node_index; /* Index of state to debug or -1. */
+static int_t debug_state_index; /* Index of state to debug or -1. */
 
 static debug_mode_t state_debug_mode; 
 /* Debug mode to use for states to debug. */
 
+static string_t analysis_file_name; /* Name of input file for analysis. */
+
 /* Analysis result output. ==================================================*/
 
 static void 
 display_result( void )
 /* Generate result file and start program to display result. */
 { 
-  string_t input;
+  char_t *input;
   string_t value_string;
   string_t count_string;
-  value_t fs;
+  value_t feat;
   int_t result_count;
 
   if (! analysis_has_nodes()) 
@@ -82,20 +86,22 @@ display_result( void )
 
   /* Print analysis results. */
   result_count = 0;
-  for (fs = first_analysis_result(); fs != NULL; fs = next_analysis_result())
+  for (feat = first_analysis_result(); 
+       feat != NULL; 
+       feat = next_analysis_result())
   { 
     result_count++;
     if (use_display) 
     { 
-      value_string = value_to_readable( fs, FALSE, -1 );
+      value_string = value_to_readable( feat, FALSE, -1 );
       fprintf( display_stream, "%d {%s}\n", result_count, value_string );
       free_mem( &value_string );
     } 
     else
     { 
       count_string = int_to_string( result_count );
-      value_string = value_to_readable( fs, FALSE, 
-					strlen( count_string ) + 2 );
+      value_string = value_to_readable( 
+	feat, FALSE, g_utf8_strlen( count_string, -1 ) + 2 );
       printf( "%s: %s\n", count_string, value_string );
       free_mem( &value_string );
       free_mem( &count_string );
@@ -117,7 +123,9 @@ display_tree( void )
 /* Display analysis tree. */
 { 
   analysis_node_t *node;
-  string_t input, node_type = 0, value_string, surf_string;
+  string_t node_type = NULL, value_string;
+  char_t *input;
+  char_t *surf_string;
 
   if (! analysis_has_nodes()) 
     complain( "No analysis started." );
@@ -156,7 +164,7 @@ display_tree( void )
     } 
     else 
       surf_string = new_string( "", NULL );
-    value_string = value_to_readable( node->link_fs, FALSE, -1 );
+    value_string = value_to_readable( node->link_feat, FALSE, -1 );
     fprintf( display_stream, "{%s} {%s} ", surf_string, value_string );
     free_mem( &value_string );
     free_mem( &surf_string );
@@ -169,7 +177,7 @@ display_tree( void )
     } 
     else 
       surf_string = new_string( "", NULL );
-    value_string = value_to_readable( node->result_fs, FALSE, -1 );
+    value_string = value_to_readable( node->result_feat, FALSE, -1 );
     fprintf( display_stream, "{%s} {%s} ", surf_string, value_string );
     free_mem( &value_string );
     free_mem( &surf_string );
@@ -190,7 +198,7 @@ display_after_analysis( void )
 { 
   if (auto_result) 
     display_result();
-  if (auto_tree) 
+  if (auto_tree && use_display) 
     display_tree();
 }
 
@@ -222,6 +230,8 @@ do_tree( string_t arguments )
   parse_end( &arguments );
   if (! analysis_has_nodes()) 
     complain( "No analysis started." );
+  if (! use_display)
+    complain( "Can only show a tree when using a display process." );
   display_tree();
 }
 
@@ -254,7 +264,7 @@ analyse_argument( grammar_t grammar, string_t arguments )
     encode_hangul( &analysis_input );
   }
   debug_state = NULL;
-  analyse( grammar, analysis_input, BUILD_TREE, ANALYSE_ALL );
+  analyse( grammar, analysis_input, TRUE, TRUE );
 }
 
 /*---------------------------------------------------------------------------*/
@@ -267,30 +277,46 @@ analyse_line( grammar_t grammar, string_t arguments )
 {
   int_t line_number, current_line_number;
   FILE *input_stream;
-  string_t file_name, input_line;
+  char_t *input_line;
   
   input_stream = NULL;
   TRY 
   { 
     /* Read arguments. */
-    file_name = parse_absolute_path( &arguments, NULL );
-    line_number = parse_int( &arguments );
+    line_number = parse_cardinal( &arguments );
+    if (*arguments != EOS)
+    {
+      free_mem( &analysis_file_name );
+      analysis_file_name = parse_absolute_path( &arguments, NULL );
+    }
     parse_end( &arguments );
 
+    if (analysis_file_name == NULL)
+      complain( "Missing input file name." );
+
     /* Read the line from input. */
-    input_stream = open_stream( file_name, "r" );
-    free_mem( &file_name );
+    input_stream = open_stream( analysis_file_name, "r" );
     current_line_number = 0;
     input_line = NULL;
-    do 
-    { 
-      check_user_break();
-      free_mem( &input_line );
-      input_line = read_line( input_stream );
-      if (input_line == NULL) 
-	complain( "No line %d.", line_number );
-      current_line_number++;
-    } while (current_line_number < line_number);
+
+    TRY
+    {
+      do 
+      { 
+	check_user_break();
+	free_mem( &input_line );
+	current_line_number++;
+	input_line = read_line( input_stream );
+      } while (current_line_number < line_number && input_line != NULL);
+    }
+    IF_ERROR
+      print_text( error_text, " (line %d)", current_line_number );
+    END_TRY;
+    if (input_line == NULL) 
+    {
+      complain( "\"%s\" contains only %d lines.", 
+		name_in_path( analysis_file_name ), current_line_number - 1 );
+    }
     preprocess_input( input_line, FALSE );
     if (*input_line == EOS) 
       complain( "Line %d is empty.", line_number );
@@ -303,7 +329,7 @@ analyse_line( grammar_t grammar, string_t arguments )
   END_TRY;
 
   debug_state = NULL;
-  analyse( grammar, analysis_input, BUILD_TREE, ANALYSE_ALL );
+  analyse( grammar, analysis_input, TRUE, TRUE );
 }
 
 /*---------------------------------------------------------------------------*/
@@ -367,7 +393,8 @@ static command_t ma_line_command =
   "ma-line mal", do_ma_line,
   "Analyse a line in a file morphologically.\n"
   "Usage:\n"
-  "  ma-line FILE LINE -- Analyse LINE in FILE.\n"
+  "  ma-line LINE FILE -- Analyse LINE in FILE.\n"
+  "If FILE is omitted, the previous file name is used.\n"
   "\"ma-line\" can't be used in debug mode.\n"
 };
 
@@ -388,7 +415,8 @@ static command_t sa_line_command =
   "sa-line sal", do_sa_line,
   "Analyse a line in a file syntactically.\n"
   "Usage:\n"
-  "  sa-line FILE LINE -- Analyse LINE in FILE.\n"
+  "  sa-line LINE [FILE] -- Analyse LINE in FILE.\n"
+  "If FILE is omitted, the previous file name is used.\n"
   "\"sa-line\" can't be used in debug mode.\n"
 };
 
@@ -398,13 +426,14 @@ static void analyse_input( grammar_t grammar, string_t input )
 {
   value_t value;
   int_t i;
-  string_t string, count, input_readable;
+  string_t string, count;
+  char_t *input_readable;
 
   analysis_input = new_string( input, NULL );
   input_readable = new_string_readable( input, NULL );
   preprocess_input( analysis_input, FALSE );
   encode_hangul( &analysis_input );
-  analyse( grammar, analysis_input, BUILD_TREE, ANALYSE_ALL );
+  analyse( grammar, analysis_input, TRUE, TRUE );
 
   if (analysis_has_results()) 
   {
@@ -416,7 +445,8 @@ static void analyse_input( grammar_t grammar, string_t input )
     {
       i++;
       count = int_to_string( i );
-      string = value_to_readable( value, FALSE, strlen( count ) + 2) ;
+      string = value_to_readable( value, FALSE, 
+				  g_utf8_strlen( count, -1 ) + 2) ;
       printf( "\n%s: %s\n", count, string );
       free_mem( &string );
       free_mem( &count );
@@ -433,58 +463,40 @@ static void
 display_where( void )
 /* Print rule name, left and right surface. */
 { 
-  string_t surf, file, rule;
+  char_t *surf;
+  string_t file, rule;
   int_t line;
 
   source_of_instr( executed_rule_sys, pc, &line, &file, &rule );
-  printf( "At \"%s\", line %d, rule \"%s\", ", 
+  printf( "At \"%s\", line %d, rule \"%s\".\n", 
 	  name_in_path( file ), line, rule );
 
-  /* Print state's surface. */
+  /* Print state number and state's surface. */
   surf = get_surface( STATE_SURFACE );
   decode_hangul( &surf );
-  printf( "surf: %s", surf );
+  if (current_state != -1) 
+    printf( "State: %d, surf: %s", current_state, surf );
+  else
+    printf( "Surf: %s", surf );
   free_mem( &surf );
 
   /* Print link's surface. */
   surf = get_surface( LINK_SURFACE );
-  decode_hangul( &surf );
-  if (*surf != EOS) 
+  if (surf != NULL)
+  {
+    decode_hangul( &surf );
     printf( ", link: %s", surf );
-  free_mem( &surf );
+    free_mem( &surf );
+  }
 
-  /* Print state number. */
-  if (current_state != -1) 
-    printf( ", state: %d", current_state );
   printf( ".\n" );
+
   if (in_emacs_malaga_mode) 
     printf( "SHOW \"%s\":%d:0\n", file, line );
 }
 
 /*---------------------------------------------------------------------------*/
 
-static bool_t 
-analysis_node_exists( int_t node_index )
-/* Return TRUE iff an analysis node with index NODE_INDEX has been
- * generated in the last analysis. */
-{ 
-  analysis_node_t *node;
-  int_t index;
-
-  for (node = get_first_analysis_node(); 
-       node != NULL; 
-       node = get_next_analysis_node()) 
-  { 
-    index = node->index;
-    free_analysis_node( &node );
-    if (index == node_index) 
-      return TRUE;
-  }
-  return FALSE;
-}
-
-/*---------------------------------------------------------------------------*/
-
 static void 
 malaga_debug_state( int_t state, bool_t enter )
 /* Callback function for "analyse". 
@@ -493,7 +505,7 @@ malaga_debug_state( int_t state, bool_t enter )
  * It is called with ENTER == FALSE when successor rules for state with 
  * analysis node INDEX have been executed. */
 { 
-  if (state != debug_analysis_node_index) 
+  if (state != debug_state_index) 
     return;
   if (enter)
     set_debug_mode( state_debug_mode, rule_system[ top_grammar ] );
@@ -523,18 +535,18 @@ do_debug_state( string_t arguments )
   assert_not_in_debug_mode();
   check_interactive_analysis();
 
-  debug_analysis_node_index = parse_int( &arguments );
+  debug_state_index = parse_int( &arguments );
   parse_end( &arguments );
 
-  if (! analysis_node_exists( debug_analysis_node_index )) 
+  if (debug_state_index < 0 || debug_state_index >= state_count) 
     complain( "State not found." );
-  
+
   state_debug_mode = WALK_MODE;
   debug_state = malaga_debug_state;
   /* Debug mode is set by "malaga_debug_state" before and after 
    * rule application. */
 
-  analyse( top_grammar, analysis_input, BUILD_TREE, ANALYSE_ALL );
+  analyse( top_grammar, analysis_input, TRUE, TRUE       );
 }
 
 static command_t debug_state_command = 
@@ -614,7 +626,8 @@ static command_t debug_ma_line_command =
   "Analyse a line in a file morphologically.\n"
   "Execute morphology combination rules in debug mode.\n"
   "Usage:\n"
-  "  debug-ma-line FILE LINE -- Analyse LINE in FILE.\n"
+  "  debug-ma-line LINE [FILE] -- Analyse LINE in FILE.\n"
+  "If FILE is omitted, the previous file name is used.\n"
   "Rule execution stops at the first statement.\n"
   "\"debug-ma-line\" can't be used in debug mode.\n"
 };
@@ -638,7 +651,8 @@ static command_t debug_sa_line_command =
   "Analyse a line in a file syntactically.\n"
   "Execute syntax combination rules in debug mode.\n"
   "Usage:\n"
-  "  debug-sa-line FILE LINE_NO -- Analyse LINE_NO in FILE.\n"
+  "  debug-sa-line LINE [FILE]  -- Analyse LINE in FILE.\n"
+  "If FILE is omitted, the previous file name is used.\n"
   "Rule execution stops at the first statement.\n"
   "\"debug-sa-line\" can't be used in debug mode.\n"
 };
@@ -650,10 +664,11 @@ write_output( string_t input, int_t line_number, string_t error_message,
               FILE *output )
 /* Write the result of the last analysis on OUTPUT. */
 { 
-  string_t input_string, line_number_string, state_count_string, buffer;
+  string_t line_number_string, state_count_string, buffer;
   string_t value_string, result_number_string;
+  char_t *input_string;
   int_t result_count;
-  value_t fs;
+  value_t feat;
 
   input_string = new_string_readable( input, NULL );
   decode_hangul( &input_string );
@@ -688,12 +703,12 @@ write_output( string_t input, int_t line_number, string_t error_message,
     if (result_as_list)
     {
       result_count = 0;
-      for (fs = first_analysis_result();
-	   fs != NULL;
-	   fs = next_analysis_result())
+      for (feat = first_analysis_result();
+	   feat != NULL;
+	   feat = next_analysis_result())
       {
 	result_count++;
-	push_value( fs );
+	push_value( feat );
       }
       build_list( result_count );
       value_string = value_to_readable( value_stack[ --top], FALSE, -1 );
@@ -708,15 +723,15 @@ write_output( string_t input, int_t line_number, string_t error_message,
     {
       /* Print real results. */
       result_count = 0;
-      for (fs = first_analysis_result(); 
-	   fs != NULL; 
-	   fs = next_analysis_result())
+      for (feat = first_analysis_result(); 
+	   feat != NULL; 
+	   feat = next_analysis_result())
       { 
 	result_count++;
 	if (*result_format != EOS) 
 	{ 
 	  result_number_string = int_to_string( result_count );
-	  value_string = value_to_readable( fs, FALSE, -1 );
+	  value_string = value_to_readable( feat, FALSE, -1 );
 	  buffer = replace_arguments( result_format, "slrfn", 
 				      input_string, line_number_string, 
 				      result_number_string, value_string,
@@ -745,7 +760,7 @@ analyse_stream( grammar_t grammar,
 		FILE *statistics,
 		bool_t expect_quotes )
 /* Analyse words or sentences in INPUT, write result to OUTPUT.
- * Use GRAMMAR for analysis, LIST_FILE_NAME is name of INPUT stream.
+ * Use GRAMMAR for analysis.
  * If EXPECT_QUOTES == TRUE, expect quoted input lines and remove the quotes.
  * Write statistic information to STATISTICS. */
 { 
@@ -753,9 +768,9 @@ analyse_stream( grammar_t grammar,
   volatile int_t combi_recognised, robust_recognised, errors, line_number;
   volatile string_t item, error_message;
   time_t start_time, stop_time;
-  string_t input_line;
+  char_t *input_line;
   bool_t old_in_emacs_malaga_mode;
-  value_t fs;
+  value_t feat;
   double time_diff;
 
   item = (grammar == MORPHOLOGY ? "wordform" : "sentence");
@@ -787,16 +802,16 @@ analyse_stream( grammar_t grammar,
 	{ 
 	  analyses++;
           encode_hangul( &input_line );
-          analyse( grammar, input_line, NO_TREE, ANALYSE_ALL );
+          analyse( grammar, input_line, FALSE, TRUE );
           if (recognised_by_combi_rules) 
 	    combi_recognised++;
           if (recognised_by_robust_rule) 
 	    robust_recognised++;
           if (analysis_has_results()) 
 	    recognised++;
-          for (fs = first_analysis_result(); 
-	       fs != NULL; 
-               fs = next_analysis_result()) 
+          for (feat = first_analysis_result(); 
+	       feat != NULL; 
+               feat = next_analysis_result()) 
 	  { 
 	    results++; 
 	  }
@@ -877,28 +892,34 @@ analyse_file( string_t arguments, grammar_t grammar )
  * or sentence list, analyse all its lines according to GRAMMAR,
  * and write the results to a file with extension ".out". */
 { 
-  string_t list_file_name, result_file_name;
+  string_t result_file_name;
   FILE *input_stream, *output_stream;
 
   input_stream = output_stream = NULL;
-  list_file_name = result_file_name = NULL;
+  result_file_name = NULL;
   TRY 
   { 
-    list_file_name = parse_absolute_path( &arguments, NULL );
+    if (*arguments != EOS)
+    {
+      free_mem( &analysis_file_name );
+      analysis_file_name = parse_absolute_path( &arguments, NULL );
+    }
+    if (*analysis_file_name == NULL)
+      complain( "Missing input file name." );
+
     if (*arguments != EOS) 
       result_file_name = parse_absolute_path( &arguments, NULL );
     else 
-      result_file_name = concat_strings( list_file_name, ".out", NULL );
+      result_file_name = concat_strings( analysis_file_name, ".out", NULL );
     parse_end( &arguments );
-    input_stream = open_stream( list_file_name, "r" );
+    input_stream = open_stream( analysis_file_name, "r" );
     output_stream = open_stream( result_file_name, "w" );
     analyse_stream( grammar, input_stream, output_stream, stdout, FALSE );
   } 
   FINALLY 
   { 
-    close_stream( &input_stream, list_file_name );
+    close_stream( &input_stream, analysis_file_name );
     close_stream( &output_stream, result_file_name );
-    free_mem( &list_file_name );
     free_mem( &result_file_name );
   } 
   END_TRY;
@@ -918,9 +939,10 @@ static command_t ma_file_command =
 { 
   "ma-file maf", do_ma_file,
   "Analyse a word list file.\n"
-  "Usage: ma-file INPUT_FILE [OUTPUT_FILE]\n"
+  "Usage: ma-file [INPUT_FILE [OUTPUT_FILE]]\n"
   "INPUT_FILE must contain one word form on each line.\n"
   "The results are written to \"OUTPUT_FILE\".\n"
+  "If INPUT_FILE is missing, the previous file name is used.\n"
   "If OUTPUT_FILE is missing, they are written to \"INPUT_FILE.out\".\n"
   "\"ma-file\" can't be used in debug mode.\n"
 };
@@ -941,9 +963,10 @@ static command_t sa_file_command =
 { 
   "sa-file saf", do_sa_file,
   "Analyse a sentence list file.\n"
-  "Usage: sa-file INPUT_FILE [OUTPUT_FILE]\n"
+  "Usage: sa-file [INPUT_FILE [OUTPUT_FILE]]\n"
   "INPUT_FILE must contain one sentence on each line.\n"
   "The results are written to \"OUTPUT_FILE\".\n"
+  "If INPUT_FILE is missing, the previous file name is used.\n"
   "If OUTPUT_FILE is missing, they are written to \"INPUT_FILE.out\".\n"
   "\"sa-file\" can't be used in debug mode.\n"
 };
@@ -1115,6 +1138,7 @@ main( int argc, char *argv[] )
   stop_display_process();
   terminate_malaga();
   free_mem( &analysis_input );
+  free_mem( &analysis_file_name );
   free_mem( &project_file );
   terminate_basic();
   return 0;
