@@ -1,8 +1,9 @@
+;; Copyright (C) 1995 Gerald Schueller.
 ;; Copyright (C) 1995 Bjoern Beutel.
 
 (provide 'malaga)
 
-(defconst malaga-version "6.13" "Version of Malaga.")
+(defconst malaga-version "7.8" "Version of Malaga.")
 
 ;; malaga mode (for editing malaga files). ====================================
 
@@ -42,8 +43,6 @@ Commands:
 
 Variables controlling malaga-mode for all buffers:
   malaga-indent-level
-  malaga-brace-level
-  malaga-tab-always-indent
   malaga-change-to-project-directory"
   
   (interactive)
@@ -106,12 +105,12 @@ Variables controlling malaga-mode for all buffers:
 ;; Data Structures for malaga mode. ===========================================
 
 (defconst malaga-keywords-list
-  '("accept" "allo_rule" "and" "assert" "break" "choose" "combi_rule" "default"
-    "define" "else" "elseif" "end" "end_rule" "error" "foreach" "greater" 
-    "greater_equal" "if" "in" "include" "initial" "input_filter" "less" 
-    "less_equal" "matches" "not" "or" "output_filter" "parallel" "pruning_rule"
-    "repeat" "require" "result" "return" "robust_rule" "rules" "stop" "subrule"
-    "then" "while")
+  '("accept" "allo_rule" "and" "assert" "break" "choose" "combi_rule" 
+    "continue" "default" "define" "else" "elseif" "end" "end_rule" "error" 
+    "foreach" "greater" "greater_equal" "if" "in" "include" "initial" 
+    "input_filter" "less" "less_equal" "matches" "not" "or" "output_filter" 
+    "parallel" "pruning_rule" "repeat" "require" "result" "return" 
+    "robust_rule" "rules" "select" "stop" "subrule" "then" "while")
   "The keywords used in malaga rule files (as a list).")
 
 (defvar malaga-keywords nil 
@@ -126,14 +125,15 @@ Variables controlling malaga-mode for all buffers:
 	  "\\|output_filter\\|pruning_rule\\|robust_rule\\|subrule\\)\\>"))
 (defconst malaga-begin-block-re 
   (concat "\\(" malaga-rule-re "[\t ]*\\({\\|" malaga-ident-re "[\t ]*(\\)\\)"
-	  "\\|\\<\\(foreach\\|if\\|parallel\\|repeat\\)\\>[\t ]*[^\t ;]"))
+	  "\\|\\<\\(foreach\\|if\\|parallel\\|repeat\\|select\\)\\>"
+	  "[\t ]*[^\t ;]"))
 (defconst malaga-end-block-re "\\<end\\>")
-(defconst malaga-sub-block-re "\\<\\(and\\|else\\|elseif\\|while\\)\\>")
+(defconst malaga-sub-block-re "\\<\\(and\\|else\\|elseif\\|or\\|while\\)\\>")
 (defconst malaga-noindent-re 
-  "\\<\\(and\\|else\\|elseif\\|end\\|while\\|then\\)\\>")
+  "\\<\\(and\\|else\\|elseif\\|end\\|or\\|while\\|then\\)\\>")
 (defconst malaga-autoindent-lines-re
   (concat "\\<\\(and\\|[A-Za-z_]+rule\\|end\\|else\\|elseif\\|foreach\\|if"
-	  "\\|of\\|parallel\\|while\\)\\>"))
+	  "\\|of\\|parallel\\|select\\|while\\)\\>"))
 
 (defvar malaga-mode-syntax-table nil
   "Syntax table in use in Malaga-mode buffers.")
@@ -189,14 +189,6 @@ Variables controlling malaga-mode for all buffers:
 (defconst malaga-indent-level 2
   "*Indentation of Malaga statements with respect to containing block.")
 
-(defconst malaga-brace-level 1
-  "*Indentation of Malaga braces with respect to containing block.")
-
-(defconst malaga-tab-always-indent t
-  "*Non-nil means TAB in Malaga mode should always reindent the 
-   current line, regardless of where in the line point is when the TAB
-   command is used.")
-
 ;; Font-lock functions. =======================================================
 
 (defconst malaga-font-lock-keywords
@@ -257,7 +249,7 @@ Variables controlling malaga-mode for all buffers:
       (malaga-indent-line))
     (beginning-of-line)
     (skip-chars-forward " \t")
-    (if (> old-point (point))
+    (if (> old-point (point)) 
 	(goto-char old-point))))
 
 ;; Other malaga mode functions. ===============================================
@@ -271,9 +263,8 @@ Variables controlling malaga-mode for all buffers:
 	  state)
       (goto-char (point-min))
       (while (re-search-forward keywords nil t)
-	(if (progn
-	      (setq state (parse-partial-sexp ref-point (point)))
-	      (or (nth 3 state) (nth 4 state)))
+	(setq state (parse-partial-sexp ref-point (point)))
+	(if (or (nth 3 state) (nth 4 state))
 	    ()
 	  (setq ref-point (point))
 	  (funcall change-word -1))))))
@@ -358,18 +349,21 @@ Variables controlling malaga-mode for all buffers:
     (let* ((state (save-excursion (parse-partial-sexp (point-min) (point))))
 	   (nest 0)
 	   (par 0)
-	   (indent malaga-brace-level)
+	   (indent t)
 	   (complete nil)
 	   (type (catch 'nesting
 		   ;; Check if inside a string or parenthesis.
 		   (cond ((nth 3 state) (throw 'nesting 'string))
 			 ((nth 4 state) (throw 'nesting 'comment))
 			 ((> (car state) 0)
-			  (if (looking-at "[\]\>\)\}]") (setq indent 0))
+			  (skip-chars-forward " \t")
+			  (if (looking-at "[\]\>\)\}]") (setq indent nil))
 			  (goto-char (nth 1 state))
-			  (setq par (+ indent (current-column)))
+			  (cond (indent
+				 (forward-char)
+				 (skip-chars-forward " \t")))
+			  (setq par (current-column))
 			  (throw 'nesting 'contexp)))
-
 		   ;; Loop until correct indent is found.
 		   (while t
 		     (backward-sexp 1)
@@ -595,12 +589,12 @@ Variables controlling malaga-process-mode for all buffers:
     (set-buffer old-buffer)
 
     ;; Check for a filename-and-line number.
-    (if malaga-last-frame
-	(progn (malaga-display-line proc 
-				    (nth 0 malaga-last-frame)
-				    (nth 1 malaga-last-frame) 
-				    (nth 2 malaga-last-frame))
-	       (setq malaga-last-frame nil)))))
+    (cond (malaga-last-frame
+	   (malaga-display-line proc 
+				(nth 0 malaga-last-frame)
+				(nth 1 malaga-last-frame) 
+				(nth 2 malaga-last-frame))
+	   (setq malaga-last-frame nil)))))
 
 ;;-----------------------------------------------------------------------------
 
