@@ -1,4 +1,4 @@
-/* Copyright (C) 1995 Bjoern Beutel. */
+/* Copyright (C) 1997 Bjoern Beutel. */
 
 /* Description. =============================================================*/
 
@@ -12,6 +12,7 @@
 #include <stdlib.h>
 #include <setjmp.h>
 #include <time.h>
+#include <glib.h>
 #include "basic.h"
 #include "pools.h"
 #include "values.h"
@@ -47,28 +48,59 @@ string_t result_format, unknown_format, error_format;
 static string_t morphology_file, syntax_file, lexicon_file;
 static string_t symbol_file, extended_symbol_file;
 
+static bool_t info_in_project_file;
+/* Indicates whether we have read grammar info from the current project file.
+ * Used to insert empty lines between grammar infos from different project
+ * files. */
+
 /* Functions. ===============================================================*/
 
 static void 
-do_pruning_rule_option( string_t arguments )
-/* Enable/disable pruning-rule. */
+do_mor_pruning_option( string_t arguments )
+/* Set minimum number of states needed to call the morphology pruning rule. */
 { 
   if (*arguments == EOS) 
-  { 
-    printf( "pruning-rule: %s\n", 
-            get_analysis_option( PRUNING_RULE_OPTION ) ? "yes" : "no" );
-  } 
-  else 
-    set_analysis_option( PRUNING_RULE_OPTION, parse_yes_no( &arguments ) );
-  parse_end( &arguments );
+    printf( "mor-pruning: %d\n", mor_pruning_min );
+  else
+  {
+    if (rule_system[MORPHOLOGY]->pruning_rule == -1)
+      complain("No morphology pruning rule.");
+    mor_pruning_min = parse_int( &arguments );
+    parse_end( &arguments );
+  }
 }
 
-static command_t pruning_rule_option = 
+static command_t mor_pruning_option = 
 { 
-  "pruning-rule pruning", do_pruning_rule_option,
+  "mor-pruning", do_mor_pruning_option,
   "Usage:\n"
-  "  set pruning-rule yes -- Turn pruning-rule on.\n"
-  "  set pruning-rule no -- Turn pruning-rule off.\n"
+  "  set mor-pruning 0 -- Turn off morphology pruning.\n"
+  "  set mor-pruning N -- Call pruning rule for a minimum of N states.\n"
+};
+
+/*---------------------------------------------------------------------------*/
+
+static void 
+do_syn_pruning_option( string_t arguments )
+/* Set minimum number of states needed to call the syntax pruning rule. */
+{ 
+  if (*arguments == EOS) 
+    printf( "syn-pruning: %d\n", syn_pruning_min );
+  else 
+  {
+    if (rule_system[SYNTAX] == NULL || rule_system[SYNTAX]->pruning_rule == -1)
+      complain("No syntax pruning rule.");
+    syn_pruning_min = parse_int( &arguments );
+    parse_end( &arguments );
+  }
+}
+
+static command_t syn_pruning_option = 
+{ 
+  "syn-pruning", do_syn_pruning_option,
+  "Usage:\n"
+  "  set syn-pruning 0 -- Turn off syntax pruning.\n"
+  "  set syn-pruning N -- Call pruning rule for a minimum of N states.\n"
 };
 
 /*---------------------------------------------------------------------------*/
@@ -252,8 +284,6 @@ do_cache_size_option( string_t arguments )
   else 
   { 
     size = parse_int( &arguments );
-    if (size < 0) 
-      complain( "Don't use a negative cache size." );
     set_cache_size( size );
     set_analysis_option( CACHE_OPTION, (size > 0) );
   }
@@ -420,17 +450,26 @@ read_project_file( string_t project_file )
 /* Read the project file. */
 { 
   FILE *project_stream;
-  string_t project_line, project_line_p, argument, include_file, extension;
+  char_t *project_line;
+  string_t project_line_p, argument, include_file, extension;
   string_t *name_p;
   volatile int_t line_count;
   static bool_t err_pos_printed;
 
-  err_pos_printed = FALSE;
+  info_in_project_file = err_pos_printed = FALSE;
   project_stream = open_stream( project_file, "r" );
   line_count = 0;
   while (TRUE) 
   { 
-    project_line = read_line( project_stream );
+    TRY
+      project_line = read_line( project_stream );
+    IF_ERROR
+    {
+      print_text( error_text, " (\"%s\", line %d)",
+		  name_in_path( project_file ), line_count + 1 );
+      err_pos_printed = TRUE;
+    }
+    END_TRY;
     if (project_line == NULL) 
       break;
     line_count++;
@@ -479,16 +518,13 @@ read_project_file( string_t project_file )
 	} 
 	else if (strcmp_no_case( argument, "info:" ) == 0) 
         { 
+	  /* Insert an empty line if we already have info that stems from a
+	   * different project file. */
+	  if (grammar_info->string_size > 0 && ! info_in_project_file)
+	    add_char_to_text( grammar_info, '\n' );
 	  add_to_text( grammar_info, project_line_p );
 	  add_char_to_text( grammar_info, '\n' );
-	}
-	else if (strcmp_no_case( argument, "char-set:" ) == 0
-		 || strcmp_no_case( argument, "char_set:" ) == 0)
-	{
-	  if (char_set != NULL) 
-	    complain( "Char set already defined." );
-	  char_set = parse_word( &project_line_p );
-	  parse_end( &project_line_p );
+	  info_in_project_file = TRUE;
 	}
 	free_mem( &argument );
       
@@ -518,6 +554,7 @@ read_project_file( string_t project_file )
     free_mem( &project_line );
   }
   close_stream( &project_stream, project_file );
+  info_in_project_file = FALSE;
 }
 
 /*---------------------------------------------------------------------------*/
@@ -528,11 +565,11 @@ static command_t *malaga_options[] =
   &alias_option, &auto_result_option, &auto_tree_option, 
   &auto_variables_option, &cache_size_option, &display_line_option, 
   &error_format_option, &hidden_option, &mor_incomplete_option, 
-  &mor_out_filter_option, &pruning_rule_option, &result_format_option,
-  &result_list_option, &robust_rule_option, &sort_records_option, 
-  &switch_option, &syn_incomplete_option, &syn_in_filter_option, 
-  &syn_out_filter_option, &transmit_line_option, &unknown_format_option, 
-  &use_display_option, &use_ksc_option,
+  &mor_out_filter_option, &mor_pruning_option, &result_format_option,
+  &result_list_option, &robust_rule_option, &roman_hangul_option, 
+  &sort_records_option, &switch_option, &syn_incomplete_option, 
+  &syn_in_filter_option, &syn_out_filter_option, &syn_pruning_option,
+  &transmit_line_option, &unknown_format_option, &use_display_option, 
   NULL
 };
 
@@ -554,8 +591,6 @@ init_malaga( string_t project_file )
 	      project_file );
   }
   read_project_file( project_file );
-  if (char_set == NULL) 
-    char_set = new_string( "iso8859-1", NULL );
   if (morphology_file == NULL) 
     complain( "Missing morphology rules." );
   if (lexicon_file == NULL) 
@@ -630,7 +665,6 @@ terminate_malaga( void )
   terminate_values();
   free_text( &grammar_info );
   terminate_input();
-  free_mem( &char_set );
 }
 
 /* End of file. =============================================================*/
