@@ -13,7 +13,7 @@
 #include <stdlib.h>
 #include <setjmp.h>
 #include <time.h>
-#include <locale.h>
+#include <glib.h>
 #include "basic.h"
 #include "pools.h"
 #include "values.h"
@@ -25,6 +25,7 @@
 #include "lex_compiler.h"
 #include "input.h"
 #include "commands.h"
+#include "commands_interactive.h"
 #include "options.h"
 #include "breakpoints.h"
 #include "debugger.h"
@@ -43,7 +44,7 @@ static bool_t lex_tree_to_output = FALSE;
 static string_t lexicon_file, rule_file, symbol_file, project_file;
 static string_t prelex_file;
 
-static string_t base_fs_string; /* Last base feature structure. */
+static string_t base_feat_string; /* Last base feature structure. */
 
 /* Functions. ===============================================================*/
 
@@ -69,6 +70,20 @@ display_where( void )
 
 /*---------------------------------------------------------------------------*/
 
+static void
+read_lexicon_file_name( string_t *arguments )
+{
+  if (**arguments != EOS)
+  {
+    free_mem( &lexicon_file );
+    lexicon_file = parse_absolute_path( arguments, NULL );
+  }
+  if (lexicon_file == NULL)
+    complain( "Missing lexicon file name." );
+}
+
+/*---------------------------------------------------------------------------*/
+
 static void 
 display_result( void )
 /* Display result in the modes that have been switched on after analysis. */
@@ -77,12 +92,12 @@ display_result( void )
   { 
     start_display_process();
     fprintf( display_stream, "allomorph\n" );
-    print_lex_tree( display_stream, "%n %s {%f}" );
+    print_lex_buffer( display_stream, "%n %s {%f}" );
     fprintf( display_stream, "end\n" );
     fflush( display_stream );
   } 
   else 
-    print_lex_tree( stdout, NULL );
+    print_lex_buffer( stdout, NULL );
 }
 
 /*---------------------------------------------------------------------------*/
@@ -110,20 +125,18 @@ static void
 do_read_constants( string_t arguments )
 /* Read the constants in the lexicon with name on line ARGUMENTS. */
 { 
-  string_t lexicon_name;
-
   assert_not_in_debug_mode();
-  lexicon_name = parse_absolute_path( &arguments, NULL );
+  read_lexicon_file_name( &arguments );
   parse_end( &arguments );
-  read_lex_constants( lexicon_name );
-  free_mem( &lexicon_name );
+  read_lex_constants( lexicon_file );
 }
 
 static command_t read_constants_command = 
 { 
   "read-constants", do_read_constants,
   "Read the constants from the definitions in a lexicon file.\n"
-  "Usage: read-constants LEXICON_FILE\n"
+  "Usage: read-constants [LEXICON_FILE]\n"
+  "If LEXICON_FILE is omitted, the previous lexicon file name is used.\n"
   "\"read-constants\" can't be used in debug mode.\n"
 };
 
@@ -134,37 +147,37 @@ do_ga_file( string_t arguments )
 /* Generate allomorphs of a base lexicon.
  * Write the allomorphs readably into file. */
 { 
-  string_t lexicon_name, output_name;
+  string_t output_name;
   FILE *output_stream;
 
   assert_not_in_debug_mode();
 
   output_stream = NULL;
-  output_name = lexicon_name = NULL;
+  output_name = NULL;
   TRY 
   { 
-    lexicon_name = parse_absolute_path( &arguments, NULL );
+    read_lexicon_file_name( &arguments );
     if (*arguments != EOS) 
       output_name = parse_absolute_path( &arguments, NULL );
-    else 
-      output_name = concat_strings( lexicon_name, ".out", NULL );
     parse_end( &arguments );
 
     set_debug_mode( RUN_MODE, NULL );
     lex_tree_to_output = FALSE;
-    generate_allos_for_file( lexicon_name, NULL, TRUE );
+    generate_allos_for_file( lexicon_file, NULL, TRUE );
     lex_tree_to_output = TRUE;
 
-    output_stream = open_stream( output_name, "w" );
-    print_lex_tree( output_stream, allo_format );
-    print_lex_statistics( stdout );
-    close_stream( &output_stream, output_name );
+    if (output_name != NULL)
+    {
+      output_stream = open_stream( output_name, "w" );
+      print_lex_buffer( output_stream, allo_format );
+      print_lex_statistics( stdout );
+      close_stream( &output_stream, output_name );
+    }
   } 
   FINALLY 
   { 
     close_stream( &output_stream, NULL );
     free_mem( &output_name );
-    free_mem( &lexicon_name );
   }
   END_TRY;
 }
@@ -173,9 +186,10 @@ static command_t ga_file_command =
 { 
   "ga-file gaf", do_ga_file,
   "Generate allomorphs from the entries in a lexicon file.\n"
-  "Usage: ga-file LEXICON_FILE [ALLO_FILE]\n"
-  "The results are written to \"ALLO_FILE\".\n"
-  "If ALLO_FILE is missing, they are written to \"LEXICON_FILE.out\".\n"
+  "Usage: ga-file [LEXICON_FILE [ALLO_FILE]]\n"
+  "If LEXICON_FILE is omitted, the previous lexicon file name is used.\n"
+  "If ALLO_FILE is given, the results are written to \"ALLO_FILE\".\n"
+  "The results can also be displayed by the command \"result\".\n"
   "\"ga-file\" can't be used in debug mode.\n"
 };
 
@@ -186,22 +200,13 @@ do_debug_ga_file( string_t arguments )
 /* Generate allomorphs of the base lexicon with name in ARGUMENTS.
  * Execute rules in debug mode. */
 { 
-  string_t lexicon_name;
-
   assert_not_in_debug_mode();
-  lexicon_name = NULL;
-  TRY 
-  { 
-    lexicon_name = parse_absolute_path( &arguments, NULL );
-    parse_end( &arguments );
-    set_debug_mode( WALK_MODE, allo_rule_sys );
-    lex_tree_to_output = FALSE;
-    generate_allos_for_file( lexicon_name, NULL, TRUE );
-    lex_tree_to_output = TRUE;
-  }
-  FINALLY 
-    free_mem( &lexicon_name );
-  END_TRY;
+  read_lexicon_file_name( &arguments );
+  parse_end( &arguments );
+  set_debug_mode( WALK_MODE, allo_rule_sys );
+  lex_tree_to_output = FALSE;
+  generate_allos_for_file( lexicon_file, NULL, TRUE );
+  lex_tree_to_output = TRUE;
 }
 
 static command_t debug_ga_file_command = 
@@ -209,7 +214,8 @@ static command_t debug_ga_file_command =
   "debug-ga-file dgaf", do_debug_ga_file,
   "Generate allomorphs from the entries in a lexicon file.\n"
   "Execute the rules in debug mode.\n"
-  "Usage: debug-ga-file LEXICON_FILE\n"
+  "Usage: debug-ga-file [LEXICON_FILE]\n"
+  "If LEXICON_FILE is omitted, the previous lexicon file name is used.\n"
   "\"debug-ga-file\" can't be used in debug mode.\n"
 };
 
@@ -220,16 +226,14 @@ generate_allomorphs_for_line( string_t arguments )
 /* Generate allomorphs for ARGUMENTS, which should consist
  * of a file name and a line number. */
 { 
-  string_t lexicon_name;
   int_t line;
 
-  lexicon_name = parse_absolute_path( &arguments, NULL );
-  line = parse_int( &arguments );
+  line = parse_cardinal( &arguments );
+  read_lexicon_file_name( &arguments );
   parse_end( &arguments );
   lex_tree_to_output = FALSE;
-  generate_allos_for_line( lexicon_name, line );
+  generate_allos_for_line( lexicon_file, line );
   lex_tree_to_output = TRUE;
-  free_mem( &lexicon_name );
 }
 
 /*---------------------------------------------------------------------------*/
@@ -249,8 +253,9 @@ static command_t ga_line_command =
 { 
   "ga-line gal", do_ga_line,
   "Generate allomorphs from a single entry in a file.\n"
-  "Usage: ga-line FILE LINE\n"
+  "Usage: ga-line LINE [FILE]\n"
   "The first lexicon entry at or behind LINE in FILE is read in.\n"
+  "If FILE is omitted, the previous file name is used.\n"
   "\"ga-line\" can't be used in debug mode.\n"
 };
 
@@ -271,9 +276,10 @@ static command_t debug_ga_line_command =
   "debug-ga-line dgal", do_debug_ga_line,
   "Generate allomorphs from a single entry in a file.\n"
   "Execute allomorph rules in debug mode.\n"
-  "Usage: debug-ga-line FILE LINE\n"
+  "Usage: debug-ga-line LINE [FILE]\n"
   "The first lexicon entry at or behind LINE in FILE is read in.\n"
   "Allomorph rule execution stops at the first statement.\n"
+  "If FILE is omitted, the previous file name is used.\n"
   "\"debug-line\" can't be used in debug mode.\n"
 };
 
@@ -286,16 +292,16 @@ generate_allomorphs( string_t arguments )
   /* If no argument given, re-analyze last argument */
   if (*arguments == EOS) 
   { 
-    if (base_fs_string == NULL) 
+    if (base_feat_string == NULL) 
       complain( "No previous base feature structure." );
   } 
   else 
   { 
-    free_mem( &base_fs_string );
-    base_fs_string = new_string( arguments, NULL );
+    free_mem( &base_feat_string );
+    base_feat_string = new_string( arguments, NULL );
   }
   lex_tree_to_output = FALSE;
-  generate_allos_for_string( base_fs_string );
+  generate_allos_for_string( base_feat_string );
   lex_tree_to_output = TRUE;
 }
 
@@ -316,9 +322,9 @@ static command_t ga_command =
   "ga", do_ga,
   "Generate allomorphs from a feature structure argument.\n"
   "Usage:\n"
-  "  ga FS -- Generate allomorphs for feature structure FS.\n"
+  "  ga FEAT -- Generate allomorphs for feature structure FEAT.\n"
   "  ga -- Re-generate allomorphs for the last argument.\n"
-  "The allomorphs are printed on screen.\n"
+  "The allomorphs are shown on screen.\n"
   "\"ga\" can't be used in debug mode.\n"
 };
 
@@ -340,10 +346,10 @@ static command_t debug_ga_command =
   "Generate allomorphs from the feature structure argument. "
   "Execute allomorph rules in debug mode.\n"
   "Usage:\n"
-  "  debug-ga FS -- Generate allomorphs for feature structure FS.\n"
+  "  debug-ga FEAT -- Generate allomorphs for feature structure FEAT.\n"
   "  debug-ga -- Re-generate allomorphs for the last argument.\n"
   "Rule execution stops at the first statement.\n"
-  "The allomorphs are printed on screen.\n"
+  "The allomorphs are shown on screen.\n"
   "\"debug-ga\" can't be used in debug mode.\n"
 };
 
@@ -385,8 +391,9 @@ static command_t allo_format_option =
 static command_t *mallex_options[] = 
 { 
   &alias_option, &allo_format_option, &auto_variables_option, 
-  &display_line_option, &hidden_option, &sort_records_option, &switch_option, 
-  &transmit_line_option, &use_display_option, &use_ksc_option,
+  &display_line_option, &hidden_option, &roman_hangul_option, 
+  &sort_records_option, &switch_option, &transmit_line_option, 
+  &use_display_option,
   NULL
 };
 
@@ -411,7 +418,8 @@ read_project_file( string_t file_name )
 { 
   FILE *project_stream;
   string_t include_file;
-  string_t project_line, project_line_p, argument, extension;
+  string_t project_line_p, argument, extension;
+  char_t *project_line;
   string_t *name_p;
   volatile bool_t binary = FALSE;
   volatile int_t line_count;
@@ -422,7 +430,15 @@ read_project_file( string_t file_name )
   line_count = 0;
   while (TRUE) 
   { 
-    project_line = read_line( project_stream );
+    TRY
+      project_line = read_line( project_stream );
+    IF_ERROR
+    {
+      print_text( error_text, " (\"%s\", line %d)",
+		  name_in_path( file_name ), line_count + 1 );
+      err_pos_printed = TRUE;
+    }
+    END_TRY;
     if (project_line == NULL) 
       break;
     line_count++;
@@ -470,14 +486,6 @@ read_project_file( string_t file_name )
 	  read_project_file( include_file );
 	  free_mem( &include_file );
 	}
-	else if (strcmp_no_case( argument, "char-set:" ) == 0
-		 || strcmp_no_case( argument, "char_set:" ) == 0)
-	{
-	  if (char_set != NULL) 
-	    complain( "Char set already defined." );
-	  char_set = parse_word( &project_line_p );
-	  parse_end( &project_line_p );
-	}
 	free_mem( &argument );
 
 	if (name_p != NULL && *name_p == NULL && *project_line_p != EOS) 
@@ -599,8 +607,6 @@ main( int argc, char *argv[] )
   }
   if (project_file != NULL) 
     read_project_file( project_file );
-  if (char_set == NULL) 
-    char_set = new_string( "iso8859-1", NULL );
   if (rule_file == NULL) 
     complain( "Missing allomorph rule file name." );
   if (symbol_file == NULL) 
@@ -660,12 +666,12 @@ main( int argc, char *argv[] )
     {
     case TEXT_MODE:
       generate_allos_for_file( lexicon_file, NULL, TRUE );
-      print_lex_tree( stdout, allo_format );
+      print_lex_buffer( stdout, allo_format );
       break;
     case BINARY_MODE:
       generate_allos_for_file( lexicon_file, prelex_file, TRUE );
       set_binary_file_name( &object_file, lexicon_file );
-      write_lex_tree( object_file );
+      write_lex_buffer( object_file );
       free_mem( &object_file );
       break;
     case PRELEX_MODE:
@@ -683,7 +689,7 @@ main( int argc, char *argv[] )
   }
 
   free_aliases();
-  free_mem( &base_fs_string );
+  free_mem( &base_feat_string );
   free_mem( &allo_format );
   stop_display_process();
   terminate_lex_compiler();
@@ -698,7 +704,6 @@ main( int argc, char *argv[] )
   free_mem( &symbol_file );
   free_mem( &lexicon_file );
   free_mem( &project_file );
-  free_mem( &char_set );
   terminate_input();
   terminate_basic();
   return 0;
